/* Simulation model of a barber shop. Main class is Simulation.

   The shop opens at 8:00 in the morning, and closes at 18:00.
   Customers enter roughly every 20 minutes, and serving them takes
   roughly 30 minutes. When the shop closes, everybody in the shop will
   still be served.

   When a customer enters and there are already customers waiting,
   the customer may chose to leave and come back another day.

   The simulation uses the notion of event driven simulation, where
   a priority queue of events is ordered in increasing event time.
   The event scheduler fetches the earliest event, sets the model
   time to this event's time, and calls the run method of the process
   for which the event was scheduled. The process performs the action
   scheduled for the event, and may schedule later events in doing so.

   As a half-realistic assumption, erlang distributions are assumed
   for both the customer arrival rate, and the barber's service time.
*/

import java.util.Random;
import java.util.Queue;
import java.util.LinkedList;

/* Simultation events: pairs of event time and process that
   should perform some action when the event occurs.
   To provide a total order, there is integer instance ID
   as a tie breaker in it also. */

class Event implements Comparable{
    static int instanceCounter;
    double time;
    int uniqueId;
    Process process;

    Event(double time, Process process) {
	this.time = time;
	this.process = process;
	uniqueId = instanceCounter++;
    }

    // equals does not need to be implemented: only identical events compare equal

    public int compareTo(Object other){
	Event e2 = (Event)other;
	if (time < e2.time) return -1;
	if (time > e2.time) return +1;
	if (uniqueId < e2.uniqueId) return -1;
	if (uniqueId > e2.uniqueId) return +1;
	return 0;
    }
}

/* Main class of the event-driven simulation. A scheduler maintains
   a list of events. In its run method, it continues processing the
   earliest event until the model runs out of events. */
class Scheduler {
    Queue<Object> pqueue = new java.util.PriorityQueue<Object>();

    public Random random = new Random();

    /* Time is represent in minutes since midnight. */
    double currentTime;
    public double getCurrentTime() {
	return currentTime;
    }

    public String timeAsString(){
	double rest = currentTime;
	int hours = (int)(currentTime/60);
	rest -= 60*hours;
	int minutes = (int)(rest);
	rest -= minutes;
	int seconds = (int)(rest*60);
	return hours+":"+minutes+":"+seconds;
    }

    /* Main API for processes to schedule events - delay is relative
       to the current time (and ought to be positive. */
    public void wait(Process p, double delay) {
	schedule(p, currentTime+delay);
    }

    /* Advanced scheduling API, with absolute time. */
    public void schedule(Process p, double time){
	pqueue.add(new Event(time, p));
    }

    /* Main event loop. */
    public void run()
    {
	System.out.println("Simulation starts");
	while(pqueue.size() > 0) {
	    Event e = (Event)pqueue.remove();
	    currentTime = e.time;
	    System.out.println("Current time: "+timeAsString());
	    e.process.run(this);
	}
	System.out.println("Simulation stopped at "+timeAsString());
    }

    /* Java's Random class doesn't provide erlang distributions, so
       we do our own here. */
    public double nextErlang(int a)
    {
        double product = 1.0;
        for(int i=1; i<=a; i++)
            product *= random.nextDouble();

        return -Math.log(product);
    }
}

/* Interface to be implemented by all activities in the model. */
interface Process {
    void run(Scheduler s);
}

/* Main activity. Processes customers as long as they are
   waiting; becomes idle when there are no customers anymore.
   An event is scheduled if:
   - there is a customer waiting (timeout 0), or
   - there is a customer being served right now (timeout:
   completion of service). */
class BarberShop implements Process{

    Queue<Object> customers = new LinkedList<Object>();
    Object currentCustomer;
    int served, rejected;
 
    public boolean idle() {
	return currentCustomer == null;
    }

    public void run(Scheduler s){
	if (idle()) {
	    currentCustomer = customers.remove();
	    System.out.println("Start serving "+currentCustomer);
	    // Assumption: service time follows an erlang distribution, 
	    // with an average service time of 30.
	    // Realistically, a lower bound > 0 should be used
	    s.wait(this, s.nextErlang(30));
	} else {
	    System.out.println("Done with "+currentCustomer);
	    currentCustomer = null;
	    served++;
	    if (customers.size() > 0)
		// proceed with the next customer without any break
		s.wait(this, 0);
	}
    }

    /* A customer enters the shop. If too many people are waiting
       already, it choses to leave. Otherwise, it joins the queue.
       If the barber is currently idle, schedule an event for immediate
       start of service. */
    public void enter(Scheduler s, Object o){
	System.out.println("Customer "+o+" entering at "+s.timeAsString()+" already waiting "+customers.size());
	if (customers.size() > 4){
	    // if there are more than four customers, 
	    // there is a 50-50 chance that the customer will leave
	    if (s.random.nextBoolean()) {
		rejected++;
		return;
	    }
	}
	customers.add(o);
	if (idle())
	    // can start serving right away
	    s.wait(this, 0);
    }
}

/* Simple customer type - customers are only identified by number. */
class Customer{
    static int counter;
    int id;
    public Customer(){
	id = ++counter;
    }
    public String toString(){
	return "Customer "+id;
    }
}

/* Process producing customers. The main class makes the process
   start at 8:00; it voluntarily ends at 18:00. 
*/
class City implements Process{
    // Assume there is a single barber shop in our city
    BarberShop shop = new BarberShop();

    public void run(Scheduler s) {
	// we produce new customers between 8:00 and 18:00
	// roughly every 20 minutes
	if (s.getCurrentTime() < 18*60) {
	    shop.enter(s, new Customer());
	    s.wait(this, s.nextErlang(20));
	}
    }
}

class Simulation{
    /* Main function. Create a scheduler, and the city,
       which in turn creates the shop. Let the city start at 8:00,
       then wait for the simulation to complete. */
    public static void main(String args[]){
	Scheduler s = new Scheduler();
	City c = new City();
	// live starts at 8:00
	s.wait(c, 8*60);
	// now run the simulation
	s.run();
	System.out.println("served: "+c.shop.served+" rejected: "+c.shop.rejected);
    }
}
	
    